/*
 *  Cryptographic "Data-as-Key" Feedback Analyser
 *
 *  Copyright (C) 2005, 2008 Luke Kenneth Casson Leighton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*#define TEST_HISTOGRAM*/

#define K_SZ 16
#define D_SZ 16

#ifdef TEST_RIJNDAEL
#include "aes.h"
#define ALG "rjd"
#endif

#ifdef TEST_SERPENT
#include "serpent.h"
#define ALG "serpent"
#endif

#ifdef TEST_DES
#include <openssl/des.h>
#define ALG "des"
#undef K_SZ
#undef D_SZ
#define K_SZ 7
#define D_SZ 8

static void
str_to_key(const unsigned char *str, unsigned char *key)
{
	int i;

	key[0] = str[0]>>1;
	key[1] = ((str[0]&0x01)<<6) | (str[1]>>2);
	key[2] = ((str[1]&0x03)<<5) | (str[2]>>3);
	key[3] = ((str[2]&0x07)<<4) | (str[3]>>4);
	key[4] = ((str[3]&0x0F)<<3) | (str[4]>>5);
	key[5] = ((str[4]&0x1F)<<2) | (str[5]>>6);
	key[6] = ((str[5]&0x3F)<<1) | (str[6]>>7);
	key[7] = str[6]&0x7F;
	for (i=0;i<8;i++) {
		key[i] = (key[i]<<1);
	}
}

#endif

#ifdef TEST_BF
#include <openssl/blowfish.h>
#define ALG "blowfish"
#endif

#ifdef TEST_CAST
#include <openssl/cast.h>
#define ALG "cast"
#endif

#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>

#include "util.h"


#define sqrt2 1.41421356237309504880

/*
 * the basic premise of this program USED TO BE as follows:
 *
 * for a given initial key "k" and initial data "buf",
 * repeat num_tests times:
 *
 * 	o = E(k, buf); k = o; buf = xor(buf, o)
 *
 * if the algorithm is any good, then using the key as the
 * data for the next block will show no signs of non-randomness
 * in the output.  this just seems like good common sense.
 *
 * (and Rijndael/128 shows problems on blocks as quickly as 5-10 loops)
 *
 * signs of non-randomness aren't such a big hairy deal... as long as
 * there's no correlation between bits being set in the initial key
 * and bits being set in the initial data... and that's what the
 * histogram is there to find out.
 *
 * use gnuplot in weird and wonderful ways to look at the plots in 3d...
 *
 */

static int do_decrypt = 0;

static double pval_bitbucket[D_SZ*8][K_SZ*8];

extern double erfc(double x);

static void encrypt(const unsigned char *key, const unsigned char *buf,
					unsigned char *out, int decrypt)
{
	/* o = E(k, buf); k = o; buf = xor(buf, o) */
#ifdef TEST_RIJNDAEL
	aes_context ctx;
	aes_set_key( &ctx, (unsigned char*)key, 128 );
	if (decrypt)
		aes_decrypt( &ctx, (unsigned char*)buf, out );
	else
		aes_encrypt( &ctx, (unsigned char*)buf, out );
#endif
#ifdef TEST_SERPENT
	unsigned int ctx[140];
	serpent_set_key( ctx, key, 128 );
	serpent_encrypt( &ctx, buf, out);
#endif
#ifdef TEST_CAST
	CAST_KEY ctx;
	CAST_set_key( &ctx, K_SZ, key );
	CAST_ecb_encrypt( &buf[0], &out[0], &ctx, CAST_ENCRYPT);
	CAST_ecb_encrypt( &buf[8], &out[8], &ctx, CAST_ENCRYPT);
#endif
#ifdef TEST_BF
	BF_KEY ctx;
	BF_set_key( &ctx, K_SZ, key );
	BF_ecb_encrypt( &buf[0], &out[0], &ctx, BF_ENCRYPT);
	BF_ecb_encrypt( &buf[8], &out[8], &ctx, BF_ENCRYPT);
#endif
#ifdef TEST_DES
	DES_key_schedule ctx;
	unsigned char key_data[8];
	str_to_key(key, key_data);
	DES_set_key_unchecked( &key_data, &ctx );
	if (decrypt)
		DES_ecb_encrypt( buf, out, &ctx, DES_DECRYPT );
	else
		DES_ecb_encrypt( buf, out, &ctx, DES_ENCRYPT );
	/*print_data(out, 8);*/
#endif
}

/* 
 * o1 = E(k xor xorkey, d xor xordata)
 * o2 = E(k           , d)
 * o = o1 xor o2
 *
 * count num bits in o
 *
 */
int analyse(  unsigned char *baseline,
		unsigned char *d, unsigned char *k, unsigned char *xordata,
		unsigned char *xorkey,
		unsigned char *out)
{
	unsigned char buf[D_SZ];
	/*unsigned char out[D_SZ];*/
	unsigned char key[K_SZ];
	unsigned char test_data[D_SZ]; /* make this 32..64... for larger tests */
	double pval;

	int num_fails = 0;

	memcpy(key, k, K_SZ);
	memcpy(buf, d, D_SZ);

	xor(key, key, xorkey);
	xor(buf, buf, xordata);
	encrypt(key, buf, out, do_decrypt);

	xor(test_data, out, baseline);

	/* count the number of bits, create a p-value */
	pval = data_test(test_data, sizeof(test_data));

	if (pval < 0.01)
		num_fails++;

	/*printf("testval: %10.f pval: %f\n", testval, pval);*/
	return num_fails;
}

struct datapoint
{
	unsigned char i;
	unsigned char j;
	int bitbucket;
	struct datapoint *next;
};

static int bitbucket[D_SZ*8][K_SZ*8];
static struct datapoint *sortedbucket = NULL;

void dump_bitbucket(void)
{
	fwrite(bitbucket, sizeof(bitbucket), 1, stdout);
}
			

void print_bitbucket(int num_samples)
{
	int i, j;
	for (i = 0; i < D_SZ*8; i++)
	{
		int print = 0;
		for (j = 0; j < K_SZ*8; j++)
		{
			if (bitbucket[i][j] != 0)
			{
				print = 1;
				break;
			}
		}
		if (!print)
			continue;

		for (j = 0; j < D_SZ*8; j++)
		{
			if (bitbucket[i][j] != 0)
				printf("%3d ", (int)(100.*D_SZ*8.*K_SZ*8.*(double)bitbucket[i][j]/(double)num_samples));
		}
		printf("\n");
	}
}


int test(struct datapoint *i1, struct datapoint *i2)
{
	return 1;
	if (i1->bitbucket > i2->bitbucket)
		return 1;
	/*if (i1->bitbucket == i2->bitbucket && i1->j > i2->j)
		return 1;*/
	return 0;
}

void insert_datapoint(int i, int j, int bitbucket)
{
	struct datapoint *d = malloc(sizeof(struct datapoint));
	struct datapoint *start = sortedbucket;

	d->i = i;
	d->j = j;
	d->bitbucket = bitbucket;
	d->next = NULL;

	if (start == NULL)
	{
		sortedbucket = d;
		return;
	}

	if (test(sortedbucket, d))
	{
		d->next = sortedbucket;
		sortedbucket = d;
		return;
	}

	while (start->next != NULL)
	{
		if (test(start->next, d))
		{
			d->next = start->next;
			start->next = d;
			return;
		}
		start = start->next;
	}
	start->next = d;
}

static unsigned char rnumsj[128];
static unsigned char rnumsi[128];

static void remove_old_graph(const char *alg_name, int samples)
{
	char fname[256];

	sprintf(fname, "%s.%d.txt", alg_name, samples);
	unlink(fname);
}

static void write_floatbucket(double bb[D_SZ*8][K_SZ*8],
		const char *alg_name, int samples)
{
	int i, j;
	char fname[256];
	FILE *f;

	sprintf(fname, "%s.%d.txt", alg_name, samples);
	f = fopen(fname, "w");

	for (i = 0; i < D_SZ*8; i++)
	{
		for (j = 0; j < K_SZ*8; j++)
		{
			fprintf(f, "%d.\t%d.\t%g.\n", i, j, bb[i][j]);
		}
	}
	fclose(f);
}

static void write_bucket(int bb[D_SZ*8][K_SZ*8],
		const char *alg_name, int samples)
{
	int i, j;
	char fname[256];
	FILE *f;

	sprintf(fname, "%s.%d.txt", alg_name, samples);
	f = fopen(fname, "w");

	for (i = 0; i < D_SZ*8; i++)
	{
		for (j = 0; j < K_SZ*8; j++)
		{
			fprintf(f, "%d.\t%d.\t%d.\n", i, j, bb[i][j]);
		}
	}
	fclose(f);
}

static void print_sortedbucket(const char *alg_name, int samples)
{
	write_bucket(bitbucket, alg_name, samples);
}

void swap(unsigned char *rnums, unsigned char i, unsigned char j)
{
	unsigned char tmp = rnums[i];
	rnums[i] = rnums[j];
	rnums[j] = tmp;
}

void init_rnums(unsigned char *rnums)
{
	unsigned char i;
	for (i = 0; i < 128; i++)
		rnums[i] = i;
}

void init_ij_rnums(void)
{
	init_rnums(rnumsj);
	init_rnums(rnumsi);
}

void print_ij_rnums(void)
{
	unsigned char i;
	for (i = 0; i < 128; i++)
		printf("%3d %3d\n", rnumsi[i], rnumsj[i]);
}

void diffcount( int *icount, int *jcount)
{
	struct datapoint *d;
	d = sortedbucket;
	*jcount = 0;
	*icount = 0;
	while (d != NULL)
	{
		int idiff = abs(d->j - rnumsi[d->i]);
		int jdiff = abs(d->i - rnumsj[d->j]);

		*icount += idiff;
		*jcount += jdiff;

		d = d->next;
	}
}

#if 0

static int prevdiffi = 1<<30;
static int prevdiffj = 1<<30;

void get_best_diffcount(void)
{
	int counti = 0;
	int countj = 0;

	int r = rand();

	int i_j = (r>>0) & 127;
	int i_i = (r>>7) & 127;
	int j_j = (r>>14) & 127;
	int j_i = (r>>21) & 127;

	static int rnums_inited;
	
	if (!rnums_inited)
	{
		init_ij_rnums();
		rnums_inited = 1;
	}

	swap(rnumsi, i_i, i_j);
	swap(rnumsj, j_i, j_j);

	diffcount(&counti, &countj);

	if (countj >= prevdiffj)
	{
		swap(rnumsj, j_i, j_j);
	}
	else
	{
		prevdiffj = countj;
	}
	if (counti >= prevdiffi)
	{
		swap(rnumsi, i_i, i_j);
	}
	else
	{
		prevdiffi = counti;
	}

}

void repeat_best_count(int goes)
{
	int i;
	for (i = 0; i < goes; i++)
		get_best_diffcount();
}
#endif

extern int gen_rand_impl(unsigned char *, unsigned int len);


#define RECORD_OUT 128

int main( int ac, char *av[] )
{
	int k, i, j;
	unsigned char xorbuf[D_SZ];
	unsigned char key[K_SZ];
	unsigned char xorkey[K_SZ];
	unsigned char input[D_SZ];
	unsigned char out[D_SZ];
	unsigned char baseline[D_SZ];
	float keyprob[K_SZ*8];
	char fn[128];
	char fname[128];
	char c;

	int total = 0;
	int max_run = 1024;
	int interactive = 0;
	int encrypt_only = 0;
	int read_key = 0;
	int interactive_run = 1024;
	int oldk;
	int output_baseline;

	int double_enc_keybit = -1;
	int double_enc_databit = -1;

	int i1;

	int key_set[K_SZ*8];
	int bl_set[128];

	strcpy(fname, ALG);

	for (i = 0; i < D_SZ*8; i++)
		for (j = 0; j < K_SZ*8; j++)
			pval_bitbucket[i][j] = 0.0;

    while( (c = getopt( ac, av, "K:D:l:bkdeio:")) != EOF )
	{
		switch(c)
		{
			case 'k':
			{
				read_key = 1;
				break;
			}
			case 'b':
			{
				output_baseline = 1;
				break;
			}
			case 'd':
			{
				do_decrypt = 1;
				break;
			}
			case 'e':
			{
				output_baseline = 1;
				encrypt_only = 1;
				break;
			}
			case 'i':
			{
				interactive = 1;
				break;
			}
			case 'o':
			{
				strncpy(fname, optarg, sizeof(fname));
				break;
			}
			case 'D':
			{
				double_enc_databit = atoi(optarg);
				break;
			}
			case 'K':
			{
				double_enc_keybit = atoi(optarg);
				break;
			}
			case 'l':
			{
				interactive_run = atoi(optarg);
				break;
			}
		}
	}

	sprintf(fn, "%s-pval", fname);

	if (interactive)
	{
		max_run = interactive_run;
	}

	if (interactive && !read_key)
	{
		for (i = 0; i < K_SZ*8; i++)
		{
			fscanf(stdin, "%f\n", &keyprob[i]);
			printf("%d %f\n", i, keyprob[i]);
			fflush(stdout);
		}
	}

#if 0
	for (i = 0; i < K_SZ*8; i++)
	{
		/* count the number of bits, create a p-value, blah blah */
		int count = i;
		double testval = (count - ((double)K_SZ*4)) * 2;
		double sobs = fabs(testval) / sqrtl(K_SZ*8);
		double f = sobs / sqrt2;
		double pval = erfc(f);
		/*p_buckets[i] = pval;*/

		printf("%d %.30f\n", i, pval);
	}
	exit(0);
#endif

	chdir("graphs");

	for (k = 0; k < max_run; k++) /* the more, the merrier */
	{
		int block_total_num_fails = 0;
		if (interactive)
		{
			if (read_key)
			{
				read_input(key, K_SZ);
			}
			else
			{
				weighted_pseud(key, keyprob);
			}
			read_input(input, D_SZ);
			printf("read input\n");
			fflush(stdout);
		}
		else
		{
			gen_rand_impl(key, K_SZ);  /*paranoia: get a decent PRNG, dude */
			gen_rand_impl(input, D_SZ);  /*paranoia: get a decent PRNG, dude */
		}
#if 0
		memset( input, 0x0, D_SZ );
		/*bit_set(input, k);*/
		(*(int*)input) = k;
#endif

		if (double_enc_keybit != -1 && double_enc_databit != -1)
		{
			unsigned char key2[K_SZ];
			unsigned char o2[D_SZ];

			memcpy(key2, key, K_SZ);
			bit_set(key2, double_enc_keybit);
			memcpy(o2, input, D_SZ);
			bit_set(o2, double_enc_databit);

			/* O = D(K^kb, D^bb)
			 * test k0-127,d0-127 E(K^kn, O^dn) against D
			 */
			memcpy(baseline, input, D_SZ); /* input BEFORE rev-enc is baseline*/
			encrypt(key2, o2, input, !do_decrypt); /* reverse encrypt */
			if (output_baseline)
			{
				write_data(input, D_SZ); /* encrypted data is in 'input' */
				printf("encrypt\n");
				fflush(stdout);
			}
		}
		else
		{
			encrypt(key, input, baseline, do_decrypt);
			if (output_baseline)
			{
				write_data(baseline, D_SZ);
				printf("encrypt\n");
				fflush(stdout);
			}
		}

		if (encrypt_only)
		{
			continue;
		}
			
		for (i1 = 0; i1 < K_SZ * 8; i1++)
		{
			key_set[i1] = bit_is_set(key, i1);
		}

		for (i1 = 0; i1 < D_SZ * 8; i1++)
		{
			bl_set[i1] = bit_is_set(baseline, i1);
		}

		for (i = 0; i < D_SZ * 8; i++)
		{
			memset( xorbuf, 0x0, D_SZ );
			bit_set(xorbuf, i);
			/*bit_set(buf, k % 128);*/

			/*printf("%d-> ", i);*/
			for (j = 0; j < K_SZ * 8; j++)
			{
				int num_fails;
				memset( xorkey, 0x0, K_SZ );
				bit_set(xorkey, j);

				/* get the number of times that the
				 * p-value (see NIST SP800-22b.pdf)
				 * fell belowout2 0.01 */
				num_fails = analyse(baseline, input, key, xorbuf, xorkey, out);
				/* create a histogram of the failures, using
				 * the key bit index and the data bit index */
				bitbucket[i][j] += num_fails;
				total += num_fails;
				block_total_num_fails += num_fails;

				if (num_fails)
				{
					if (bit_is_set(out, i) == bit_is_set(key, j))
						pval_bitbucket[i][j] += 1;
				}
				if (0)
				{
					double pval_out;
					unsigned char xorkd[K_SZ];
					/*xor(xorkd, out, key);*/
					pval_out = data_test(xorkd, sizeof(xorkd));

					if (pval_out < 0.005 || pval_out > 0.72)
						pval_bitbucket[i][j] += 1;
				}
			}

		}
		printf("num_fails: %d\n", block_total_num_fails);
		fprintf(stderr, "%d\n", k);
		if ((k % RECORD_OUT != 0) || (k == 0))
		{
			continue;
		}

		if (encrypt_only)
			continue;

		oldk = k-RECORD_OUT*2;
		if (oldk > 0)
			remove_old_graph(fname, oldk);

		print_sortedbucket(fname, k);

		sprintf(fn, "%s-pval", fname);
		if (oldk > 0)
			remove_old_graph(fn, oldk);
		write_floatbucket(pval_bitbucket, fn, k);

#if 0
		for (i1 = 0; i1 < 4; i1++)
		{
			for (j1 = 0; j1 < 4; j1++)
			{
				char fn[128];
				int i2;
				int j2;
				int st[128][128];

				for (i2 = 0; i2 < 128; i2++)
					for (j2 = 0; j2 < 128; j2++)
						st[i2][j2] = stats[i1][j1][i2][j2];

				sprintf(fn, "%s-%d.%d", ALG, i1, j1);
				/*if (oldk > 0)
					remove_old_graph(fn, oldk);*/
				write_bucket(st, fn, max_run);

			}
		}
#endif
	}
	/*print_bitbucket(total);*/
	/*print_sortedbucket(total);*/
	/*dump_bitbucket();*/
	if (!encrypt_only)
	{
		print_sortedbucket(fname, k);
		write_floatbucket(pval_bitbucket, fn, k);
	}
#if 0
	repeat_best_count(10000);
	print_sortedbucket(total);
	print_ij_rnums();
#endif
	printf("done\n");
	fflush(stdout);
	return( 0 );
}

